#include <Adafruit_NeoPixel.h>
#ifdef __AVR__
  #include <avr/power.h>
#endif

#define PIN 6

#define NUM_LEDS 70

#define MAX_BRIGHTNESS 255

Adafruit_NeoPixel strip_a = Adafruit_NeoPixel(NUM_LEDS, PIN, NEO_GRBW + NEO_KHZ800);
Adafruit_NeoPixel strip_b = Adafruit_NeoPixel(NUM_LEDS, PIN-1, NEO_GRBW + NEO_KHZ800);

byte neopix_gamma[] = {
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,
    2,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,  4,  4,  5,  5,  5,
    5,  6,  6,  6,  6,  7,  7,  7,  7,  8,  8,  8,  9,  9,  9, 10,
   10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16,
   17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22, 23, 24, 24, 25,
   25, 26, 27, 27, 28, 29, 29, 30, 31, 32, 32, 33, 34, 35, 35, 36,
   37, 38, 39, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 50,
   51, 52, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 66, 67, 68,
   69, 70, 72, 73, 74, 75, 77, 78, 79, 81, 82, 83, 85, 86, 87, 89,
   90, 92, 93, 95, 96, 98, 99,101,102,104,105,107,109,110,112,114,
  115,117,119,120,122,124,126,127,129,131,133,135,137,138,140,142,
  144,146,148,150,152,154,156,158,160,162,164,167,169,171,173,175,
  177,180,182,184,186,189,191,193,196,198,200,203,205,208,210,213,
  215,218,220,223,225,228,231,233,236,239,241,244,247,249,252,255 };

int led = 13;


// It's going to be a lot easier to just maintain pointers to
// an event call at a time period. Sounds good, yay!
struct Event {
  unsigned long timeMillis;
  void (*functionPtr)();
};

#define MAX_EVENTS 50
Event eventArray[MAX_EVENTS];
int numEvents = 0;

void setup() {
  // This is for Trinket 5V 16MHz, you can remove these three lines if you are not using a Trinket
  #if defined (__AVR_ATtiny85__)
    if (F_CPU == 16000000) clock_prescale_set(clock_div_1);
  #endif
  
  // End of trinket special code
  strip_a.setBrightness(MAX_BRIGHTNESS);
  strip_a.begin();
  strip_a.show(); // Initialize all pixels to 'off'

  strip_b.setBrightness(MAX_BRIGHTNESS);
  strip_b.begin();
  strip_b.show(); // Initialize all pixels to 'off'

  pinMode(led, OUTPUT); 
  randomSeed(analogRead(0));
  Serial.begin(115200);

  // Start the events rolling
  BassBeatOn();
  Scrub();

}

void loop() {
  //ambulance();
  //Bangarang();

  struct Event poppedEvent2;
  if (NextEventTime() < millis()) {
    poppedEvent2 = PopEvent();
    poppedEvent2.functionPtr();
  }
  //BassBeat();
  //Scrub();
  //Devolve();


  /* Degrees of freedom
    * Apparent motion
    * Changing color quickly
    * Devolve/fade


  Need some help to get the linear fade working from x to y in n steps
  */
}



unsigned long NextEventTime() {
  return eventArray[0].timeMillis;
}

void InsertEvent(struct Event event) {
  // Maintain sorted list
  struct Event tempEvent;
  struct Event eventToInsert = event;
  
  for (int i = 0; i < numEvents; i++) {
    if (eventArray[i].timeMillis <= eventToInsert.timeMillis) {
      // Event to insert is not early enough
      continue;
    }
    // Swap events, since eventToInsert is less than i'th event
    tempEvent = eventArray[i];
    eventArray[i] = eventToInsert;
    eventToInsert = tempEvent;
  }
  
  eventArray[numEvents] = eventToInsert;
  numEvents++;
  
  if (numEvents > MAX_EVENTS) {
    Error();
  }
}

// Not sure why it won't let me use Event as a return type but whatever
//Event poppedEvent;
struct Event PopEvent() {
  // Don't want to deal with circular buffers right now
  // so just migrating everything down
  Event toReturn;
  toReturn = eventArray[0];
  //toReturn.timeMillis = eventArray[0].timeMillis;
  //toReturn.functionPtr = eventArray[0].functionPtr;
  for (int i = 1; i < numEvents; i++) {
    eventArray[i-1] = eventArray[i];
  }
  numEvents--;
  return toReturn;
}

void Error() {
  // Error condition
  for(uint16_t i=0; i< strip_a.numPixels(); i+= 2) {
    strip_a.setPixelColor(i, strip_a.Color(255, 0, 0, 0));
    strip_b.setPixelColor(i, strip_b.Color(255, 0, 0, 0));
  }
  strip_a.show();
  strip_b.show();
  while(1);
}



void ClearAll() {
  for(uint16_t i=0; i< strip_a.numPixels(); i+= 1) {
    strip_a.setPixelColor(i, strip_a.Color(0, 0, 0, 0));
    strip_b.setPixelColor(i, strip_b.Color(0, 0, 0, 0));
  }
}


uint8_t bassTrackBeats[4] = {MAX_BRIGHTNESS/8, MAX_BRIGHTNESS, MAX_BRIGHTNESS/8,MAX_BRIGHTNESS};
                              
int bassTrackBeatsIndex = 0;
// https://www.all8.com/tools/bpm.htm
int beatsPerMinute = 108;
int millisPerBeat = 555;//(60*1000)/beatsPerMinute;

unsigned long curMillis = 0;
unsigned long prevMillis = 0;
unsigned long millisIlluminated = 50;
int onState = 0;
uint32_t c = 0;
int timesCalled = 0;
// All lights white
void BassBeatOn() {
  InsertEvent({millis() + millisIlluminated, BassBeatOff});
  InsertEvent({millis() + millisPerBeat, BassBeatOn});
  
  for(uint16_t i=0; i<strip_a.numPixels(); i+= 2) {
    strip_a.setPixelColor(i, strip_a.Color(0, 0, 0, 5));//bassTrackBeats[bassTrackBeatsIndex]));
    strip_b.setPixelColor(i, strip_b.Color(0, 0, 0, 5));//bassTrackBeats[bassTrackBeatsIndex]));
  }
  strip_a.show();
  strip_b.show();
}
      
void BassBeatOff() {
  ClearAll();
  strip_a.show();
  strip_b.show();
  // Don't need to insert on event since it was done already
}



void Devolve() {
  // Say start at white

  for(uint16_t i=0; i< strip_a.numPixels(); i+= 2) {
  // Then devolve into random colors
    strip_a.setPixelColor(i, Wheel((int)map(i, 0, strip_a.numPixels()/2, 0, 255)));
    strip_b.setPixelColor(i, Wheel((int)map(i, 0, strip_a.numPixels()/2, 0, 255)));
  }
  strip_a.show();
  strip_b.show();
  delay(500);

}

int width = 2;
int offset = 0;
int increment = 1;
void Scrub() {
  
  if (offset > 15+15) {
    increment = -1;
  }
  if (offset < 0+15) {
    increment = 1;
  }
  
  ClearAll();
  for(uint16_t i=0+offset; i< width+offset; i+= 2) {
    strip_a.setPixelColor(i, strip_a.Color(255, 40, 0, 0));
    strip_b.setPixelColor(i, strip_b.Color(255, 40, 0, 0));
  }
  strip_a.show();
  strip_b.show();
  offset += increment;
  InsertEvent({millis() + 5, Scrub});
  //delay(30);
}



// Fill the dots one after the other with a color
void colorWipe(uint32_t c, uint8_t wait) {
  for(uint16_t i=0; i<strip_a.numPixels(); i++) {
    strip_a.setPixelColor(i, c);
    strip_b.setPixelColor(i, c);
    strip_a.show();
    strip_b.show();
    delay(wait);
  }
}


// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
uint32_t Wheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
    return strip_a.Color(255 - WheelPos * 3, 0, WheelPos * 3,0);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
    return strip_a.Color(0, WheelPos * 3, 255 - WheelPos * 3,0);
  }
  WheelPos -= 170;
  return strip_a.Color(WheelPos * 3, 255 - WheelPos * 3, 0,0);
}


uint8_t white(uint32_t c) {
  return (c >> 24);
}
uint8_t red(uint32_t c) {
  return (c >> 16);
}
uint8_t green(uint32_t c) {
  return (c >> 8);
}
uint8_t blue(uint32_t c) {
  return (c);
}
